<?php
/**
 * ownCloud
 *
 * @author Jesus Macias Portela <jesus@owncloud.com>
 * @author Juan Pablo Villafañez Ramos <jvillafanez@owncloud.com>
 * @copyright (C) 2014 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */


namespace OCA\sharepoint\lib;

/**
 * SHAREPOINT
 *
 * PHP Class to manage Sharepoint instance as virtual file system
 *
 * @author Jesus Macias Portela
 * @version 0.1.0
 *
 */

require_once  'sharepoint/3rdparty/libSharepoint/src/sharepoint/SoapSharepointWrapper.php';
require_once  'sharepoint/3rdparty/libSharepoint/src/sharepoint/DAVSharepointWrapper.php';

class SHAREPOINT extends \OC\Files\Storage\Common {

    private static $enableLogs = FALSE;

    private $sharepointClient = NULL;
    private $sharepointDAVClient = NULL;
    private $root = '/';
    private $id;
    private $fileTree = array();
    private $treeTimestamp = NULL;
    private $spSite = NULL;
    private $user = NULL;
    private $password = NULL;
    private $mountPoint = NULL;
    private $listName = NULL;
    private $tmpFileGroup = array();

    public function __construct($parameters) {
        $this->spSite = $parameters['apiurl'];
        $this->user = $parameters['user'];
        $this->password = $parameters['password'];
        $this->mountPoint = $parameters['mountPoint'];
        $this->listName = $parameters['listName'];

        $this->sharepointClient = new \sharepoint\SoapSharepointWrapper($parameters['user'], $parameters['password'],
                                                                        $parameters['apiurl'], '2010');
        $this->sharepointDAVClient = new \sharepoint\DAVSharepointWrapper($parameters['user'], $parameters['password'],
                                                                        $parameters['apiurl']);

    }

    /**
     * Get the FS id (for ownCloud purposes)
     *
     * @return string the FS id
     */
    public function getId(){
        return 'sharepoint::' . $this->user . '@' . $this->spSite. '/' . $this->listName . '::' . $this->mountPoint;
    }

    /**
     * Create a directory. Notice that all the parent directories should exists
     *
     * @param string $path the directory we want to create
     * @return bool true if the directory is created, false otherwise
     */
    public function mkdir($path) {
        self::log('mkdir: '.$path, \OC_Log::DEBUG);
        $path = trim($path, '/');
        try{
            $response = $this->sharepointClient->createFolder($this->listName.'/'.$path);
        } catch (\Exception $e){
            self::log('EXCEPTION - mkdir - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
            return false;
        }
        if (isset($response['code']) && $response['code'] == '200'){
                $content = $this->getContent(dirname('/'.$path));
                $this->fileTree['/'.$path] = $content['/'.$path];
                return true;
            }
        return false;
    }

    /**
     * Remove the directory and its contents
     *
     * @param string $path the FS path that we want to delete
     * @return bool true if deleted, false otherwise
     */
    public function rmdir($path) {
        self::log('rmdir: '.$path, \OC_Log::DEBUG);
        $path = trim($path, '/');
        try{
            $response = $this->sharepointClient->deleteFolder($this->listName.'/'.$path);
        } catch (\Exception $e){
            self::log('EXCEPTION - rmdir - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
            return false;
        }
        if (isset($response['code']) && $response['code'] == '200') {
                $this->treeBuilder($path, true);
                return true;
            }
        return false;
    }

    /*
     * Open the directory
     *
     * @param string $path FS path
     * @return resource a directory resource
     */
    public function opendir($path) {
        self::log('opendir: '.$path, \OC_Log::DEBUG);
        $path = trim($path, '/');
        $content = $this->getDirContent($path);
        \OC\Files\Stream\Dir::register('sp:' . $this->mountPoint.$path, $content);
        return opendir('fakedir://sp:'.$this->mountPoint.$path);
    }

    /**
     * Check the path has updated in the given time. The check is usually based on the mtime
     * of the path, so if the path's filemtime is greater than $time, the file has updated
     * Under special condition (mainly the reliability of the mtime) we also check the contents
     * of the path to decide whether the path has changed or not
     *
     * @param string $path the FS path
     * @param int $time timestamp to compare against
     * @return bool true if the file has updated, false otherwise
     */
    public function hasUpdated($path, $time) {
        self::log('hasUpdated: '.$path, \OC_Log::DEBUG);
        $path = trim($path, '/');
        if ($this->treeBuilder($path) === false) {
            throw new \OCP\Files\StorageNotAvailableException();
        }

        if ($this->isRootDir($path)) {
            if($this->fileTree['/'.$path]['EPOXModified'] > $time){
                return true;
            }
        } else if($this->filetype($path) === 'dir'){
            foreach ($this->fileTree as $key => $value) {
                if($this->fileTree[$key]['EPOXModified'] > $time){
                    return true;
                }
            }
        } else if($this->filetype($path) === 'file'){
            if($this->fileTree['/'.$path]['EPOXModified'] > $time){
                return true;
            }
        }
        return false;
    }

    /**
     * @param string $path
     */
    public function filemtime($path) {
        $this->treeBuilder($path);
        $path = trim($path, '/');
        if ($this->isRootDir($path)) {
            return (isset($this->fileTree['/']['EPOXModified'])) ? $this->fileTree['/']['EPOXModified'] : 0;
        }
        if (isset($this->fileTree['/'.$path]['EPOXModified'])) {
            return $this->fileTree['/'.$path]['EPOXModified'];
        } else {
            return 0;
        }
    }

    /**
     * perform the stat call over the file
     *
     * @param string $path the FS path
     * @param array|false an array with information or false in case of error
     */
    public function stat($path) {
        self::log('stat: '.$path, \OC_Log::DEBUG);
        $path = trim($path, '/');

        $this->treeBuilder($path);

        if (!isset($this->fileTree['/'.$path])) {
            self::log('stat($path) no data returning false ', \OC_Log::DEBUG);
            return false;
        }
        self::log('stat: size '.($this->fileTree['/'.$path]['size'] != false ? (int)$this->fileTree['/'.$path]['size'] : 0), \OC_Log::DEBUG);
        self::log('stat: time '.$this->fileTree['/'.$path]['EPOXModified'], \OC_Log::DEBUG);
        $stat=array('size'  => ($this->fileTree['/'.$path]['size'] != false ? (int)$this->fileTree['/'.$path]['size'] : 0),
                    'atime' => $this->fileTree['/'.$path]['EPOXModified'],
                    'mtime' => $this->fileTree['/'.$path]['EPOXModified'],
                    'ctime' => $this->fileTree['/'.$path]['EPOXModified']);
        return $stat;
    }

    /**
     * Get the file type of the path
     *
     * @param string $path FS path
     * @return string the type of the path (usually 'dir' or 'file')
     */

    public function filetype($path) {
        self::log('filetype: '.$path, \OC_Log::DEBUG);
        $path = trim ($path, '/');
        if ($this->isRootDir($path)) {
            return 'dir';
        }
        $this->treeBuilder($path);
        if (isset($this->fileTree['/'.$path])) {
            if ($this->fileTree['/'.$path]['type'] == 1){
                self::log('filetype: '.$path.' is dir', \OC_Log::DEBUG);
                return 'dir';
            }
            else {
                self::log('filetype: '.$path.' is file', \OC_Log::DEBUG);
                return 'file';
            }
        }
        return false;
    }

    /**
     * Check if the path is readable
     *
     * @param string $path the path
     * @return bool true if it's readable, false if not
     */
    public function isReadable($path) {
        self::log('isReadable: '.$path, \OC_Log::DEBUG);
        if ($this->isRootDir($path)) {
            return true;
        }
        $this->treeBuilder($path);
        if (isset($this->fileTree['/'.$path]) && $this->checkPermissions($this->fileTree['/'.$path]['permMask'], 'isReadable')) {
            return true;
        }
        return false;
    }

    /**
     * Check if the path is updatable
     *
     * @param string $path the path
     * @return bool true if it's updatable, false if not
     */
    public function isUpdatable($path) {
        self::log('isUpdatable: '.$path, \OC_Log::DEBUG);
        if ($this->isRootDir($path)) {
            return false;
        }
        $this->treeBuilder($path);
        if(isset($this->fileTree['/'.$path]) && $this->checkPermissions($this->fileTree['/'.$path]['permMask'], 'isUpdatable')) {
            return true;
        }
        return false;
    }

    /**
     * Check if the path is creatable
     *
     * @param string $path the path
     * @return bool true if it's creatable, false if not
     */
    public function isCreatable($path) {
        self::log('isCreatable: '.$path, \OC_Log::DEBUG);
        if ($this->isRootDir($path)) {
            do{
                $u = uniqid();
            }while($this->file_exists('/'.$u));
            if($this->touch('/'.$u)){
                $this->unlink('/'.$u);
                $this->fileTree['/']['permMask'] = '0x000000000000000f';
                return true;
            }
        }
        $this->treeBuilder($path);
        if(isset($this->fileTree['/'.$path]) && $this->checkPermissions($this->fileTree['/'.$path]['permMask'], 'isCreatable')) {
            return true;
        }
        return false;
    }

    /**
     * Check if the path is deletable
     *
     * @param string $path the path
     * @return bool true if it's deletable, false if not
     */
    public function isDeletable($path) {
        self::log('isDeletable: '.$path, \OC_Log::DEBUG);
        if ($this->isRootDir($path)) {
            return false;
        }
        $this->treeBuilder($path);
        if(isset($this->fileTree['/'.$path]) && $this->checkPermissions($this->fileTree['/'.$path]['permMask'], 'isDeletable')) {
            return true;
        }
        return false;
    }

    /**
     * Check if the path is sharable
     *
     * @param string $path the path
     * @return bool true if it's sharable, false if not
     */
    public function isSharable($path) {
        $value = \OCP\Config::getAppValue('sharepoint', 'global_sharing', false);
        self::log('isSharable: '.$path.' '.$value, \OC_Log::DEBUG);
        return $value;
    }

    /**
     * Check the the file defined by path exists
     *
     * @param string $path FS path
     * @return bool true if the file exists, false otherwise
     */
    public function file_exists($path) {
        $path = trim($path, '/');
        if ($this->isRootDir($path)) {
            return true;
        }

        $this->treeBuilder($path);
        if(isset($this->fileTree['/'.$path])) {
                self::log('file_exist: '.$path. ' TRUE', \OC_Log::DEBUG);
                return true;
            }
        self::log('file_exist: '.$path. ' FALSE', \OC_Log::DEBUG);
        return false;
    }

    /**
     * Remove the path. If the path is a directory, this function falls back to the rmdir
     * function
     *
     * @param string $path FS path to be deleted
     * @return bool true if the path no longer exists, false otherwise
     */
    public function unlink($path) {
        self::log('unlink: '.$path, \OC_Log::DEBUG);
        $path = trim($path, '/');
        $this->treeBuilder($path);
        if(isset($this->fileTree['/'.$path])) {
            try{
                $response = $this->sharepointClient->deleteFile($this->listName, $this->fileTree['/'.$path]['id'], $this->listName.'/'.$path);
            } catch (\Exception $e){
                self::log('EXCEPTION - unlink - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
                return false;
            }
            if (isset($response['code']) && $response['code'] == '200') {
                $this->treeBuilder($path, true);
                return true;
            }
        }
        return false;
    }

    /**
     * Rename the files
     *
     * @param string $path1 the old name of the path
     * @param string $path2 the new name of the path
     * @return bool true if the rename is successful, false otherwise
     */
    public function rename($path1, $path2) {

        self::log('rename: '.$path1.'-'.$path2, \OC_Log::DEBUG);
        if ($this->filetype($path1) == 'dir' && strpos(trim($path2, "/"), '/') === FALSE) {
            try{
                $response = $this->sharepointClient->renameFolder($this->listName, $this->fileTree['/'.$path1]['id'], basename($path2));
            } catch (\Exception $e){
                self::log('EXCEPTION - rename - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
                return false;
            }
        }
        else if ($this->filetype($path1) == 'dir' && strpos(trim($path2, "/"), '/') !== FALSE) {
            self::log('rename: moving folder '.$path1.' to '.$path2, \OC_Log::DEBUG);
            try{
                $response = $this->sharepointDAVClient->moveFolder($this->listName, $path1, $path2);
            } catch (\Exception $e){
                self::log('EXCEPTION - rename - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
                return false;
            }
        }
        else {
            try{
                $response = $this->sharepointClient->copyFile($this->listName, $this->spSite.'/'.$this->listName.'/'.$path1, $this->spSite.'/'.$this->listName.'/'.$path2);
            } catch (\Exception $e){
                self::log('EXCEPTION - rename - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
                return false;
            }
        }
        if (isset($response['code']) && $response['code'] == '200') {
            if($this->unlink($path1)) {
                $this->fileTree = NULL;
                return true;
            }
        }
        return false;
    }

    /**
     * Open a file resouce
     *
     * @param string $path file path to be opened
     * @param string $mode open the file in this mode
     * @return resource an opened resource to read or write
     */
    public function fopen($path, $mode) {
        self::log('fopen: '.$path.' - '.$mode, \OC_Log::DEBUG);
        switch ($mode) {
            case 'r':
            case 'rb':
                if (!$this->file_exists($path)) {
                    self::log('fopen($path, $mode) unexisting file', \OC_Log::ERROR);
                    return false;
                }
                try{
                    $fp = $this->sharepointDAVClient->downloadFile($this->listName, $path);
                } catch (\Exception $e){
                    self::log('EXCEPTION - fopen - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
                    return false;
                }
                if(!$fp){
                    return false;
                }
                $tmpFile = \OC_Helper::tmpFile();
                $handle = fopen($tmpFile, 'w+');
                fwrite($handle, stream_get_contents($fp));
                fclose($handle);
                return fopen($tmpFile, 'rb');
            case 'w':
            case 'wb':
            case 'a':
            case 'ab':
            case 'r+':
            case 'w+':
            case 'wb+':
            case 'a+':
            case 'x':
            case 'x+':
            case 'c':
            case 'c+':
                $folder = \OCP\Files::tmpFolder();
                $tmpFile = $folder . '/' . basename($path);
                $this->tmpFileGroup[$tmpFile] = $path;
                \OC\Files\Stream\Close::registerCallback($tmpFile, array($this, 'writeBack'));
                return fopen('close://'.$tmpFile, $mode);
        }
        return false;
    }

    /**
     * Writes the file.
     *
     * NOTICE: This function is for a callback in the fopen function, so it's not intended to use
     * outside
     */
    public function writeBack($tmpFile) {
        self::log('writeBack: '.$tmpFile, \OC_Log::DEBUG);
        try{
            $response = $this->sharepointDAVClient->uploadFile($tmpFile, $this->spSite.'/'.$this->listName.'/'.$this->tmpFileGroup[$tmpFile]);
        } catch (\Exception $e){
            self::log('EXCEPTION - writeBack - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
            throw new \OCP\Files\NotPermittedException($e->getMessage(), $e->getCode(), $e);
        }
        if (isset($response['code']) && $response['code'] == '200') {
            $content = $this->getContent(dirname($this->tmpFileGroup[$tmpFile]));
            if (isset($content['/'.$this->tmpFileGroup[$tmpFile]])){
                $this->fileTree['/'.$this->tmpFileGroup[$tmpFile]] = $content['/'.$this->tmpFileGroup[$tmpFile]];
                return true;
            }
        }
        return false;
    }

    /**
     * "Touch" the path. This function will only create a empty file if not exists or update
     * the mtime to the latest if it exists. Setting the mtime is not supported
     *
     * @param string $path path to be "touched"
     * @param mixed $mtime null to be updated to the latest, if it's not null, it will be ignored
     * @return bool true if file exists and mtime updated
     */
    public function touch($path, $mtime=null) {
        self::log('touch: '.$path, \OC_Log::DEBUG);
        $path = trim($path, '/');
        if (!$this->file_exists('/'.$path)) {
            self::log('touch: fichero not exists '.$path, \OC_Log::DEBUG);
            $tmpFile = \OC_Helper::tmpFile();
            try{
                $response = $this->sharepointDAVClient->uploadFile($tmpFile, $this->spSite.'/'.$this->listName.'/'.$path);
            } catch (\Exception $e){
                self::log('EXCEPTION - touch - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
                return false;
            }
            if (isset($response['code']) && $response['code'] == '200') {
                $content = $this->getContent(dirname('/'.$path));
                if (isset($content['/'.$path])){
                    $this->fileTree['/'.$path] = $content['/'.$path];
                    return true;
                }
            }
        }
        return false;
    }


    public function checkConnection() {
        //Credentials are the same for soap and webdav, so only need to check one.
        try {
            $this->sharepointClient->getListCollection();
        } catch (\Exception $e) {
            self::log('EXCEPTION - checkConnection - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
            throw new \Exception($e->getMessage(), $e->getCode());
        }
        self::log('checking Connection for list '.$this->listName.': OK', \OC_Log::DEBUG);
        return true;
    }

    public function getMountPoint() {
        return $this->mountPoint;
    }

    /**
     * @param string $message
     */
    public static function log($message, $level, $from='sharepoint') {
        if(self::$enableLogs === true){
            \OC_Log::write($from, $message, $level);
        }
    }

    /**
     * @param string $path
     */
    private function isRootDir($path) {
        return ($path == '' || $path == '/' || $path == '.' || $path == '\.');
    }

    private function treeBuilder($path, $delete=FALSE) {
        if(substr($path,0,1) !== '/'){
            $path = '/'.$path;
        }
        $path = rtrim($path, '/');
        if(!isset($this->fileTree[$path]) && !$delete){
            try{
                $p = dirname(ltrim($path, '/'));
                $r1 = $this->sharepointClient->GetListsItems($this->listName, false, $p, TRUE);
                $response = $r1;
                if(!$this->isRootDir($path) && isset($response[$path]) && $response[$path]['type'] == 1){
                    $r2 = $this->sharepointClient->GetListsItems($this->listName, false, ltrim($path, '/'), TRUE);
                    $response = array_merge($r1, $r2);
                }
                //Get info from root /
                if($this->isRootDir($path)){
                    try{
                       $lists = $this->sharepointClient->GetListCollection();
                    }catch (\Exception $e){
                        self::log('EXCEPTION - treeBuilder - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
                        return false;
                    }
                    foreach ($lists as $list) {
                        if( $list['title'] === $this->listName){
                            $this->fileTree['/'] = $list;
                            break;
                        }
                    }
                    // No other way
                    if(!isset($this->fileTree['/']['permMask'])){
                        $this->fileTree['/']['permMask'] = '0x0000000000000001';
                    }
                }
            } catch (\Exception $e){
                self::log('EXCEPTION - treeBuilder - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
                return false;
            }
            //Update file tree with new data
            foreach ($response as $key => $value) {
                $this->fileTree[$key] = $response[$key];
            }
        } else if ($delete===TRUE){
            if($this->filetype($path) === 'dir'){
                foreach ($this->fileTree as $key => $value) {
                    if(strpos($key, $path) !== false){
                        unset($this->fileTree[$key]);
                    }
                }
            } else if ($this->filetype($path) === 'file'){
                unset($this->fileTree[$path]);
            }
        }
    }


    /**
     * @param string $path
     */
    private function getContent ($path){
        self::log('getContent path: '.$path, \OC_Log::DEBUG);
        $path = trim($path, '/');
        try{
            $content = $this->sharepointClient->GetListsItems($this->listName, false, $path, TRUE);
        } catch (\Exception $e){
            self::log('EXCEPTION - getContent - '.$e->getCode().' - '.$e->getMessage() , \OC_Log::DEBUG);
            return;
        }
        return $content;
    }

    /**
     * @param string $path
     */
    private function getDirContent ($path){
        if(substr($path,0,1) !== '/'){
            $path = '/'.$path;
        } else{
            $path = rtrim($path, '/');
        }
        $files = array();
        $this->treeBuilder($path);
        foreach ($this->fileTree as $key => $value) {
            if(dirname($key) === $path){
                $subItem = basename($key, '/');
                if($subItem !== '') {
                    $files[]= $subItem;
                }
            }
        }
        return $files;
    }

    /**
     * @param string $mask
     */
    private function checkPermissions($mask, $permission){
    /*
    ViewListItems: 0x0000000000000001 Allow viewing of list items in lists,
    documents in document libraries, and Web discussion comments.
    AddListItems: 0x0000000000000002 Allow addition of list items to lists,
    documents to document libraries, and Web discussion comments.
    EditListItems: 0x0000000000000004 Allow editing of list items in lists,
    documents in document libraries, Web discussion comments,
    and to customize Web part pages in document libraries.
    DeleteListItems: 0x0000000000000008 Allow deletion of list items from lists,
    documents from document libraries, and Web discussion comments.
    */
        //This array is to store the value of each permision
        $permissions = array('isReadable'=> 1, 'isCreatable'=> 2, 'isUpdatable'=> 4, 'isDeletable'=> 8);
        if(isset($permissions[$permission]) && ($permissions[$permission] & hexdec(substr($mask, -1, 1))) > 0){
            return true;
        }
        return false;
    }

}
