<?php
/**
 * ownCloud
 *
 * @author Thomas Müller <deepdiver@owncloud.com>
 * @author Krzesimir Nowak
 * @author Iago López Galeiras
 * @copyright (C) 2014 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\user_shibboleth;
use Exception;
use OC\HintException;
use OC_L10N;
use \OCP\Authentication\IApacheBackend;
use OCP\IUserBackend;
use OCP\UserInterface;

require_once __DIR__ . '/config.php';

class USER_SHIBBOLETH implements UserInterface, IApacheBackend, IUserBackend {
	/**
	 * @var \OCP\IConfig
	 */
	private $config;

	/**
	 * @var \OCP\IAppConfig
	 */
	private $appConfig;

	public function __construct() {
		$this->config = \OC::$server->getConfig();
		$this->appConfig = \OC::$server->getAppConfig();
	}

	/**
	* Check if backend implements actions
	* @param array $actions bitwise-or'ed actions
	* @returns boolean
	*
	* Returns the supported actions as int to be
	* compared with OC_USER_BACKEND_CREATE_USER etc.
	*/
	public function implementsActions($actions) {
		return (bool)((
				OC_USER_BACKEND_GET_DISPLAYNAME
				| OC_USER_BACKEND_COUNT_USERS
				| OC_USER_BACKEND_CHECK_PASSWORD)
			& $actions);
	}

	/**
	 * Sets non-shib webDAV password
	 * @param string $uid user id
	 * @param string $password new webdav password
	 * @returns bool whether setting password succeeded
	 */
	public static function setWebDAVPassword($uid, $password) {
		if (self::checkUserExists($uid)) {
			$hash = \OC::$server->getHasher()->hash($password);
			$query = \OC_DB::prepare('UPDATE `*PREFIX*shibboleth_user` SET `webdav_password` = ? WHERE LOWER(`eppn`) LIKE LOWER(?)');
			$result = $query->execute(array($hash, $uid));

			return $result ? true : false;
		}

		return false;
	}

	/**
	 * Clears non-shib webDAV password for given user
	 * @param string $uid user id
	 * @returns bool whether removing succeeded
	 */
	public static function clearWebDAVPassword($uid) {
		if (self::checkUserExists($uid)) {
			$query = \OC_DB::prepare('UPDATE `*PREFIX*shibboleth_user` SET `webdav_password` = NULL WHERE LOWER(`eppn`) LIKE LOWER(?)');
			$result = $query->execute(array($uid));

			return $result ? true : false;
		}

		return false;
	}

	/**
	 * Checks if given user has non-shib webDAV password set
	 * @param string $uid user id
	 * @returns bool whether the password is set
	 */
	public static function checkIfUserHasWebDAVPassword($uid) {
		if (self::checkUserExists($uid)) {
			$query = \OC_DB::prepare('SELECT `webdav_password` FROM `*PREFIX*shibboleth_user` WHERE LOWER(`eppn`) LIKE LOWER(?)');
			$result = $query->execute(array($uid));
			$row = $result->fetchRow();
			if (!$row || empty($row['webdav_password'])) {
				return false;
			}
			return true;
		}

		return false;
	}

	/**
	 * counts the users who did log in at least once
	 *
	 * @return int | bool
	 */
	public function countUsers() {
		$query = \OCP\DB::prepare('SELECT count(*) FROM `*PREFIX*shibboleth_user`');
		$result = $query->execute();
		if (\OCP\DB::isError($result)) {
			return false;
		}
		return $result->fetchOne();
	}

	/**
	* delete a user
	* @param string $uid The username of the user to delete
	* @returns boolean
	*
	* Deletes a user
	*/
	public function deleteUser($uid) {
		return false;
	}

	/**
	* Get a list of all users
	* @returns array with all uids
	*
	* Get a list of all users.
	*/
	public function getUsers($search = '', $limit = null, $offset = null) {
		$query = \OC_DB::prepare('SELECT `eppn` FROM `*PREFIX*shibboleth_user` WHERE LOWER(`eppn`) LIKE LOWER(?)', $limit, $offset);
		$result = $query->execute(array($search.'%'));
		$users = array();
		while ($row = $result->fetchRow()) {
			$users[] = $row['eppn'];
		}
		return $users;
	}

	/**
	* check if a user exists
	* @param string $uid the username
	* @return boolean
	*/
	private static function checkUserExists($uid) {
		$query = \OC_DB::prepare( 'SELECT * FROM `*PREFIX*shibboleth_user` WHERE LOWER(`eppn`) = LOWER(?)' );
		$result = $query->execute( array( $uid ));
		if ($result->fetchRow()) {
			return true;
		}

		return false;
	}

	/**
	* check if a user exists
	* @param string $uid the username
	* @return boolean
	*/
	public function userExists($uid) {
		return self::checkUserExists($uid);
	}

	/**
	 * get display name of the user
	 * @param string $uid user ID of the user
	 * @return string display name
	 */
	public function getDisplayName($uid) {
		$query = \OC_DB::prepare('SELECT * FROM `*PREFIX*shibboleth_user` WHERE LOWER(`eppn`) = LOWER(?)');
		$result = $query->execute( array( $uid ));
		if ($row = $result->fetchRow()) {
			if (!empty($row['displayname'])) {
				return $row['displayname'];
			}
		}

		return $uid;

	}

	/**
	 * Get a list of all display names
	 * @param string $search
	 * @param int|null $limit
	 * @param int|null $offset
	 * @returns array with  all displayNames (value) and the corresponding uids (key)
	 *
	 * Get a list of all display names and user ids.
	 */
	public function getDisplayNames($search = '', $limit = null, $offset = null) {
		$displayNames = array();
		$query = \OC_DB::prepare('SELECT `eppn`, `displayname` FROM `*PREFIX*shibboleth_user`'
		.' WHERE LOWER(`eppn`) LIKE LOWER(?) or LOWER(`displayname`) LIKE LOWER(?)', $limit, $offset);
		$result = $query->execute(array('%'.$search.'%', '%'.$search.'%'));
		while ($row = $result->fetchRow()) {
			if (!empty($row['displayname'])) {
				$displayNames[$row['eppn']] = $row['displayname'];
			} else {
				$displayNames[$row['eppn']] = $row['eppn'];
			}
		}

		return $displayNames;
	}

	/**
	 * Check if a user list is available or not
	 * @return boolean if users can be listed or not
	 */
	public function hasUserListings() {
		return true;
    }

	public function isSessionActive() {
		// Is Shibboleth authentication active?
		if (!$this->appConfig->getValue("user_shibboleth", "shibboleth_active")) {
			\OCP\Util::writeLog( 'user_shibboleth', "shibboleth not active", \OCP\Util::DEBUG);
			return false;
		}

		// we are in a shibboleth session?
		if (!isset($_SERVER[SHIB_SESSION_ID])) {
			\OCP\Util::writeLog( 'user_shibboleth', "No Shibboleth session active.", \OCP\Util::DEBUG);
			return false;
		}

		// we need the auth user to be set
		if (!isset($_SERVER["PHP_AUTH_USER"])) {
			\OCP\Util::writeLog( 'user_shibboleth', "PHP_AUTH_USER not found in _SERVER", \OCP\Util::DEBUG);
			return false;
		}

		// eduPersonPrincipalName (a globally unique user name)
		$eppn = $this->getCurrentUserId();
		if (is_null($eppn)) {
			$serverInfo = var_export($_SERVER, true);
			\OCP\Util::writeLog( 'user_shibboleth', "Failed to login because no eppn(".SHIB_EPPN.") was provided: $serverInfo", \OCP\Util::ERROR);
			throw new HintException("The IdP did not pass the eppn (".SHIB_EPPN.") properly to this service.",
				'Please contact the administrator to get this resolved.');
		}

		// user's email address
		$emailAddress = $this->getEMailAddress();

		// Display name (given name plus surname)
		$displayName = $this->getDisplay();

		// Check if user already exists
		$query = \OC_DB::prepare("SELECT * FROM `*PREFIX*shibboleth_user` WHERE `eppn` = ?");
		$result = $query->execute(array($eppn));

		// If user doesn't already exist in the DB, then create it.
		if ($result->fetchAll()) {
			$update = \OCP\DB::prepare(
				"UPDATE `*PREFIX*shibboleth_user` SET `email` = ?, `displayname` = ? WHERE `eppn` =  ?");
			$result = $update->execute(array($emailAddress, $displayName, $eppn));

			if(\OCP\DB::isError($result)) {
				\OCP\Util::writeLog( 'user_shibboleth', \OCP\DB::getErrorMessage($result), \OCP\Util::ERROR);
			}

			$this->updateEMailAddress($emailAddress, $eppn);
		} else {
			$insQuery = \OC_DB::prepare(
				"INSERT INTO `*PREFIX*shibboleth_user` (`eppn`, `email`, `displayname`, `webdav_password`) VALUES (?, ?, ?, NULL)");
			$result = $insQuery->execute(array($eppn, $emailAddress, $displayName));

			if(\OCP\DB::isError($result)) {
				\OCP\Util::writeLog( 'user_shibboleth', \OCP\DB::getErrorMessage($result), \OCP\Util::ERROR);
			}

			$this->updateEMailAddress($emailAddress, $eppn);
		}

		return true;
	}

	/**
	 * @return string
	 * JavaScript method.
	 */
	public function getLogoutAttribute() {
		// Using Shibboleth?
		if ($this->appConfig->getValue("user_shibboleth", "shibboleth_active")) {
			if (isset($_SERVER[SHIB_EPPN]) && \OC_User::isLoggedIn()) {
				return "class='shibboleth-logout'";
			}
		}

		return '';
	}

	/**
	 * Return the id of the current user
	 * @return string|null
	 */
	public function getCurrentUserId() {
		return $this->get(SHIB_EPPN);
	}

	/**
	 * @return string|null
	 */
	protected function getEMailAddress() {
		return $this->get(SHIB_EMAIL);
	}
	/**
	 * @return string|null
	 */
	protected function getDisplay() {
		return $this->get(SHIB_DISPLAY_NAME);
	}

	/**
	 * @return string|null
	 */
	protected function get($key) {
		return isset($_SERVER[$key]) ? $_SERVER[$key] : null;
	}

	public static function setTimezone() {
		if (isset($_POST['timezone-offset'])) {
			if (is_numeric($_POST['timezone-offset'])) {
				\OC::$server->getSession()->set('timezone', intval($_POST['timezone-offset']));
			}
		}
		\OCP\JSON::success();
	}

	/**
	 * @param string $emailAddress
	 * @param string $eppn
	 */
	private function updateEMailAddress($emailAddress, $eppn) {
		if (!is_null($emailAddress)) {
			$this->config->setUserValue(
				$eppn, 'settings', 'email', $emailAddress);
		}
	}

	/**
	 * Check if the password is correct
	 * @param string $uid The username
	 * @param string $password The password
	 * @return string
	 *
	 * Check if the password is correct without logging in the user
	 * returns the user id or false
	 */
	public function checkPassword($uid, $password) {
		$query = \OC_DB::prepare('SELECT `eppn`, `webdav_password` FROM `*PREFIX*shibboleth_user` WHERE LOWER(`eppn`) = LOWER(?)');
		$result = $query->execute(array($uid));

		$row = $result->fetchRow();
		if ($row) {
			$storedHash = $row['webdav_password'];
			$newHash = '';
			if (\OC::$server->getHasher()->verify($password, $storedHash, $newHash)) {
				if (!empty($newHash)) {
					self::setWebDAVPassword($uid, $password);
				}
				return $row['eppn'];
			}
		}

		return false;
	}

	/**
	 * Gets password configuration. Falls back to defaults if no
	 * config is found or it is not sensible (lengths are less
	 * than 1). The defaults are 12 for password length, 4 for
	 * group length.
	 *
	 * @return array Array with two elements, first is a password
	 * length, second is a group length
	 */
	public static function getWebDAVPasswordConfiguration() {
		$passwordDefaultLength = 12;
		$groupDefaultLength = 4;

		$passwordLength = \OCP\Config::getSystemValue('non_shib_webdav_password_length', $passwordDefaultLength);
		if ($passwordLength < 1) {
		   $passwordLength = $passwordDefaultLength;
		}

		$groupLength = \OCP\Config::getSystemValue('non_shib_webdav_password_group_length', $groupDefaultLength);
		if ($groupLength < 1) {
		   $groupLength = $groupDefaultLength;
		}

		return array($passwordLength, $groupLength);
	}

	/**
	 * AJAX handler for generating and saving a new random
	 * password for non-shibboleth WebDAV.
	 */
	public static function ajaxGetNewPassword() {
		$uid = \OCP\USER::getUser();

		\OCP\JSON::checkLoggedIn();
		\OCP\JSON::callCheck();
		\OC::$server->getSession()->close();

		$passwordSetup = \OCA\user_shibboleth\USER_SHIBBOLETH::getWebDAVPasswordConfiguration();
		$chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789';
		$count = strlen($chars);
		$result = array();
		$fullPassword = '';
		$lengthLeft = $passwordSetup[0];

		while ($lengthLeft > 0) {
			$group = '';
			$toGenerate = min($lengthLeft, $passwordSetup[1]);

			for ($i = 0; $i < $toGenerate; ++$i) {
				$index = mt_rand(0, $count - 1);
				$group .= substr($chars, $index, 1);
			}

			$result[] = $group;
			$fullPassword .= $group;
			$lengthLeft -= $toGenerate;
		}

		if (self::setWebDAVPassword($uid, $fullPassword)) {
			\OCP\JSON::success(array('data' => $result));
		} else {
			$l = OC_L10N::get('user_shibboleth');
			\OCP\JSON::error(array('data' => array('message' => $l->t('Could not save generated password'))));
		}
	}

	/**
	 * AJAX handler for revoking a password for non-shibboleth
	 * WebDAV.
	 */
	public static function ajaxRevokePassword() {
		$uid = \OCP\USER::getUser();

		\OCP\JSON::checkLoggedIn();
		\OCP\JSON::callCheck();
		\OC::$server->getSession()->close();

		if (self::clearWebDAVPassword($uid)) {
			\OCP\JSON::success();
		} else {
			$l = OC_L10N::get('user_shibboleth');
			\OCP\JSON::error(array('data' => array('message' => $l->t('Could not revoke the password'))));
		}
	}

	/**
	 * Backend name to be shown in user management
	 *
	 * @return string the name of the backend to be shown
	 */
	public function getBackendName() {
		return 'Shibboleth';
	}
}
